#include "precompile.h"

#include "entity.h"
#include "collider.h"
#include "room.h"
#include "building.h"
#include "human.h"
#include "app.h"
#include "obstacle.h"

Entity::Entity()
{

}

Entity::~Entity()
{
    ClearColliders();
}

void Entity::Initialize()
{
    xtimer_attacher.xtimer = &room->xtimer;
}

void Entity::GetAabbBox(AabbCollider& aabb_box)
{
    aabb_box.active = true;
    aabb_box.owner = this;
}

void Entity::GetCircleBox(CircleCollider& circle_box)
{
    circle_box.active = true;
    circle_box.owner = this;
    circle_box.rad = 1;
}

bool Entity::TestCollision(Entity* b)
{
    App::Instance()->perf.test_times++;
    if (b->dead) {
        return false;
    }
    for (auto& a_collider : colliders) {
        for (auto& b_collider : b->colliders) {
            if (a_collider->Intersect(b_collider)) {
                return true;
            }
        }
    }
    return false;
}

bool Entity::TestCollision(ColliderComponent* b)
{
    if (b->owner->dead) {
        return false;
    }
    for (auto& a_collider : colliders) {
        if (a_collider->Intersect(b)) {
            return true;
        }
    }
    return false;
}

bool Entity::TestCollisionEx(const a8::Vec2& aabb_pos, AabbCollider& aabb_box)
{
    for (auto& a_collider : colliders) {
        if (a_collider->IntersectEx(aabb_pos, &aabb_box)) {
            return true;
        }
    }
    return false;
}

void Entity::ClearColliders()
{
    for (auto& itr : colliders) {
        ColliderComponent* collider = itr;
        DestoryCollider(collider);
    }
    colliders.clear();
}

void Entity::BroadcastFullState()
{
    std::set<GridCell*> grid_list;
    room->grid_service.GetAllCells(grid_id, grid_list);
    for (auto& grid : grid_list) {
        for (Human* hum : grid->human_list) {
            hum->AddToNewObjects(this);
        }
    }
}

void Entity::BroadcastDeleteState()
{
    std::set<GridCell*> grid_list;
    room->grid_service.GetAllCells(grid_id, grid_list);
    for (auto& grid : grid_list) {
        for (Human* hum : grid->human_list) {
            hum->RemoveObjects(this);
        }
    }
}

void Entity::AddCollider(ColliderComponent* collider)
{
    colliders.push_back(collider);
}

bool Entity::IsClientCached(Human* hum)
{
    ObjectSyncFlags* sync_flags = hum->GetObjectSyncFlags(GetEntityUniId());
    if (sync_flags) {
        int flags = sync_flags->flags;
        if (IsDead(hum->room)) {
            if (a8::HasBitFlag(flags, kOsfIsDead)) {
                return sync_flags->last_sync_frameno > 0;
            } else {
                return false;
            }
        } else {
           if (!a8::HasBitFlag(flags, kOsfIsDead)) {
                return sync_flags->last_sync_frameno > 0;
            } else {
                return false;
            }
         }
    }
    return false;
}

bool Entity::CanClientCache(Human* hum)
{
    ObjectSyncFlags* sync_flags = hum->GetObjectSyncFlags(GetEntityUniId());
    if (sync_flags && entity_type == ET_Obstacle) {
        Obstacle* obstacle = (Obstacle*)this;
        return !obstacle->is_door;
    }
    return sync_flags != nullptr;
}

void Entity::AddClientCache(Human* hum)
{
    ObjectSyncFlags* sync_flags = hum->GetObjectSyncFlags(GetEntityUniId());
    if (sync_flags) {
        int flags = sync_flags->flags;
        if (IsDead(hum->room)) {
            a8::SetBitFlag(flags, kOsfIsDead);
        } else {
            a8::UnSetBitFlag(flags, kOsfIsDead);
        }
        sync_flags->flags = flags;
        sync_flags->last_sync_frameno = hum->room->GetFrameNo();
    }
}
