#include "precompile.h"

#include "framemaker.h"
#include "human.h"
#include "room.h"
#include "typeconvert.h"

cs::SMUpdate* FrameMaker::MakeUpdateMsg(const Human* hum)
{
    cs::SMUpdate* msg = new cs::SMUpdate;
    {
        Room* room = hum->room;
        if (room->gas_data.gas_mode == GasJump) {
            cs::MFPlane* p = msg->mutable_plane();
            TypeConvert::ToPb(room->plane.start_point, p->mutable_start_point());
            TypeConvert::ToPb(room->plane.end_point, p->mutable_end_point());
        }
        for (auto& itr : hum->new_objects) {
            if (itr != hum &&
                itr->entity_type == ET_Player &&
                a8::HasBitFlag(((Human*)itr)->status, HS_Fly)) {
                continue;
            }
            itr->FillMFObjectFull((Human*)hum, msg->add_full_objects());
        }
        for (auto& itr : hum->part_objects) {
            Entity* entity = itr;
            if (entity->dead && hum->room->frame_no - entity->dead_frameno > 10) {
                continue;
            } else {
                if (room->gas_data.gas_mode == GasJump &&
                    entity != hum &&
                    entity->entity_type == ET_Player &&
                    a8::HasBitFlag(((Human*)entity)->status, HS_Fly)) {
                    continue;
                }
            }
            if (((Human*)hum)->pos.ManhattanDistance(entity->pos) > VIEW_RANGE + 200)  {
                continue;
            }
            entity->FillMFObjectPart((Human*)hum, msg->add_part_objects());
        }
        for (auto& itr : hum->del_objects) {
            msg->add_del_objids(itr);
        }
        for (auto& itr : hum->out_objects) {
            msg->add_out_objids(itr);
        }
        for (size_t idx : hum->shots_) {
            if (idx < room->frame_event.shots_.size()) {
                auto& tuple = room->frame_event.shots_[idx];
                if (hum->CanSee(std::get<0>(tuple))) {
                    *msg->add_shots() = std::get<1>(tuple);
                }
            }
        }
        for (size_t idx : hum->bullets_) {
            if (idx < room->frame_event.bullets_.size()) {
                auto& tuple = room->frame_event.bullets_[idx];
                if (hum->CanSee(std::get<0>(tuple))) {
                    *msg->add_bullets() = std::get<1>(tuple);
                }
            }
        }
        for (size_t idx : hum->explosions_) {
            if (idx < room->frame_event.explosions_.size()) {
                auto& tuple = room->frame_event.explosions_[idx];
                if (hum->CanSee(std::get<0>(tuple))) {
                    *msg->add_explosions() = std::get<1>(tuple);
                }
            }
        }
        for (size_t idx : hum->smokes_) {
            if (idx < room->frame_event.smokes_.size()) {
                auto& tuple = room->frame_event.smokes_[idx];
                if (hum->CanSee(std::get<0>(tuple))) {
                    *msg->add_smokes() = std::get<1>(tuple);
                }
            }
        }
        for (size_t idx : hum->emotes_) {
            if (idx < room->frame_event.emotes_.size()) {
                auto& tuple = room->frame_event.emotes_[idx];
                if (hum->CanSee(std::get<0>(tuple))) {
                    *msg->add_emotes() = std::get<1>(tuple);
                }
            }
        }
        for (size_t idx : hum->chged_bullet_nums_) {
            if (idx < room->frame_event.chged_bullet_nums_.size()) {
                Human* target = room->frame_event.chged_bullet_nums_[idx];
                if (target->curr_weapon) {
                    auto p = msg->add_chged_property_list();
                    p->set_obj_id(target->entity_uniid);
                    p->set_property_type(kPropBulletNum);
                    p->set_value(target->curr_weapon->ammo);
                }
            }
        }
        for (size_t idx : hum->chged_tank_bullet_nums_) {
            if (idx < room->frame_event.chged_tank_bullet_nums_.size()) {
                Human* target = room->frame_event.chged_tank_bullet_nums_[idx];
                if (target->curr_weapon) {
                    auto p = msg->add_chged_property_list();
                    p->set_obj_id(target->entity_uniid);
                    p->set_property_type(kPropTankBulletNum);
                    p->set_value(target->tank_weapon.ammo);
                }
            }
        }
        for (size_t idx : hum->chged_tank_oil_max_) {
            if (idx < room->frame_event.chged_tank_oil_max_.size()) {
                Human* target = room->frame_event.chged_tank_oil_max_[idx];
                auto p = msg->add_chged_property_list();
                p->set_obj_id(target->entity_uniid);
                p->set_property_type(kPropTankOil);
                p->set_property_subtype(1);
                p->set_value(target->tank_oil_max);
            }
        }
        for (size_t idx : hum->chged_tank_oil_value_) {
            if (idx < room->frame_event.chged_tank_oil_value_.size()) {
                Human* target = room->frame_event.chged_tank_oil_value_[idx];
                auto p = msg->add_chged_property_list();
                p->set_obj_id(target->entity_uniid);
                p->set_property_type(kPropTankOil);
                p->set_property_subtype(0);
                p->set_value(target->tank_oil_value);
            }
        }
        for (size_t idx : hum->chged_hps_) {
            if (idx < room->frame_event.chged_hps_.size()) {
                Human* target = room->frame_event.chged_hps_[idx];
                if (hum->CanSee(target)) {
                    {
                        auto p = msg->add_chged_property_list();
                        p->set_obj_id(target->entity_uniid);
                        p->set_property_type(kPropHp);
                        p->set_value(target->GetHP());
                    }
                    {
                        auto p = msg->add_chged_property_list();
                        p->set_obj_id(target->entity_uniid);
                        p->set_property_type(kPropMaxHp);
                        p->set_value(target->GetMaxHP());
                    }
                }
            }
        }
        if (room->frame_event.airdrops_.size() > 0) {
            *msg->mutable_airdrop() = room->frame_event.airdrops_.Get(0);
        }
        if (room->gas_data.gas_mode == GasMoving) {
            msg->set_gas_progress(room->gas_data.gas_progress);
            TypeConvert::ToPb(room->gas_data.pos_old, msg->mutable_gas_pos_old());
        }
        if (room->GetFrameNo() - room->AliveCountChgFrameNo() <= 4 ||
            room->GetFrameNo() - hum->join_frameno <= 2) {
            msg->set_alive_count(room->AliveCount());
        }
    }
    return msg;
}
